package ru.dbotthepony.mc.otm.menu

import com.google.common.collect.ImmutableList
import com.mojang.datafixers.util.Pair
import it.unimi.dsi.fastutil.bytes.ByteArrayList
import it.unimi.dsi.fastutil.ints.IntArrayList
import it.unimi.dsi.fastutil.ints.IntCollection
import it.unimi.dsi.fastutil.objects.Reference2ObjectFunction
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ReferenceArrayList
import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.RegistryFriendlyByteBuf
import net.minecraft.network.protocol.common.custom.CustomPacketPayload
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.level.ServerPlayer
import net.minecraft.util.RandomSource
import net.minecraft.world.Container
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.inventory.InventoryMenu
import net.minecraft.world.inventory.MenuType
import net.minecraft.world.inventory.Slot
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.enchantment.EnchantmentEffectComponents
import net.minecraft.world.item.enchantment.EnchantmentHelper
import net.minecraft.world.level.block.entity.BlockEntity
import net.neoforged.neoforge.network.PacketDistributor
import net.neoforged.neoforge.network.handling.IPayloadContext
import ru.dbotthepony.kommons.util.Delegate
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.capability.energy.ProfiledEnergyStorage
import ru.dbotthepony.mc.otm.player.matteryPlayer
import ru.dbotthepony.mc.otm.compat.cos.cosmeticArmorSlots
import ru.dbotthepony.mc.otm.compat.curios.curiosSlots
import ru.dbotthepony.mc.otm.compat.curios.isCurioSlot
import ru.dbotthepony.mc.otm.container.IEnhancedContainer
import ru.dbotthepony.mc.otm.container.IFilteredContainerSlot
import ru.dbotthepony.mc.otm.container.computeSortedIndices
import ru.dbotthepony.mc.otm.container.sortWithIndices
import ru.dbotthepony.mc.otm.container.util.containerSlotOrNull
import ru.dbotthepony.mc.otm.core.ResourceLocation
import ru.dbotthepony.mc.otm.core.collect.ConditionalSet
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.menu.input.BooleanInputWithFeedback
import ru.dbotthepony.mc.otm.menu.widget.ProfiledLevelGaugeWidget
import ru.dbotthepony.mc.otm.network.MatteryStreamCodec
import ru.dbotthepony.mc.otm.network.MenuDataPacket
import ru.dbotthepony.mc.otm.network.SetCarriedPacket
import ru.dbotthepony.mc.otm.network.StreamCodecs
import ru.dbotthepony.mc.otm.network.decode
import ru.dbotthepony.mc.otm.network.encode
import ru.dbotthepony.mc.otm.network.nullable
import ru.dbotthepony.mc.otm.network.readByteListUnbounded
import ru.dbotthepony.mc.otm.network.syncher.SynchableGroup
import ru.dbotthepony.mc.otm.network.wrap
import ru.dbotthepony.mc.otm.network.writeByteListUnbounded
import ru.dbotthepony.mc.otm.player.IPlayerInventorySlot
import java.util.*
import java.util.function.BooleanSupplier
import java.util.function.Consumer
import java.util.function.Predicate

data class PlayerSlot<A : Slot, B : Slot>(val functional: A, val cosmetic: B? = null)

data class EquipmentSlots(
	val armorSlots: List<PlayerSlot<MatteryMenu.EquipmentMenuSlot, Slot>>,
	val curiosSlots: List<PlayerSlot<Slot, Slot>>
)

abstract class MatteryMenu(
	menuType: MenuType<*>?,
	containerId: Int,
	val inventory: Inventory,
	val tile: BlockEntity? = null
) : AbstractContainerMenu(menuType, containerId) {
	/**
	 * Server->Client synchronizer
	 */
	val mSynchronizer = SynchableGroup()
	val synchronizerRemote = mSynchronizer.Remote()
	val player: Player get() = inventory.player
	val random: RandomSource = RandomSource.create()

	private val _playerInventorySlots = ArrayList<InventorySlot>()
	private val _playerHotbarSlots = ArrayList<InventorySlot>()
	private val _playerCombinedInventorySlots = ArrayList<InventorySlot>()
	private val _exopackChargeSlots = ArrayList<MatteryMenuSlot>()

	private val playerInputs = ArrayList<PlayerInput<*>>()

	class PlayerInputPacket(val containerId: Int, val inputId: Int, val payload: ByteArrayList) : CustomPacketPayload {
		constructor(buff: FriendlyByteBuf) : this(buff.readVarInt(), buff.readVarInt(), buff.readByteListUnbounded())

		fun write(buff: FriendlyByteBuf) {
			buff.writeVarInt(containerId)
			buff.writeVarInt(inputId)
			buff.writeByteListUnbounded(payload)
		}

		fun play(context: IPayloadContext) {
			val menu = context.player().containerMenu as? MatteryMenu ?: return
			if (menu.containerId != containerId || !menu.stillValid(context.player())) return
			val input = menu.playerInputs.getOrNull(inputId) ?: return
			if (!input.test(context.player())) return
			input.invoke(input.codec.decode(context.player().registryAccess(), payload))
		}

		override fun type(): CustomPacketPayload.Type<out CustomPacketPayload> {
			return TYPE
		}

		companion object {
			val TYPE = CustomPacketPayload.Type<PlayerInputPacket>(
				ResourceLocation(
					OverdriveThatMatters.MOD_ID,
					"menu_input"
				)
			)

			val CODEC: net.minecraft.network.codec.StreamCodec<FriendlyByteBuf, PlayerInputPacket> =
				net.minecraft.network.codec.StreamCodec.ofMember(PlayerInputPacket::write, ::PlayerInputPacket)
		}
	}

	/**
	 * Client->Server input
	 */
	inner class PlayerInput<V>(val codec: MatteryStreamCodec<RegistryFriendlyByteBuf, V>, allowSpectators: Boolean = false, val handler: (V) -> Unit) : Consumer<V>, Predicate<Player> {
		val id = playerInputs.size
		var allowSpectators by mSynchronizer.boolean(allowSpectators)

		init {
			playerInputs.add(this)
		}

		private val filters = ArrayList<Predicate<Player>>()

		init {
			filters.add { this.allowSpectators || !it.isSpectator }
		}

		fun filter(predicate: Predicate<Player>): PlayerInput<V> {
			filters.add(predicate)
			return this
		}

		override fun test(player: Player): Boolean {
			return filters.all { it.test(player) }
		}

		override fun accept(value: V) {
			if (test(player)) {
				PacketDistributor.sendToServer(PlayerInputPacket(containerId, id, codec.encode(player.registryAccess(), value)))
			}
		}

		internal fun invoke(value: Any?) {
			handler.invoke(value as V)
		}
	}

	inner class SortInput(val container: Container, val settings: IItemStackSortingSettings) {
		val input = PlayerInput(MatteryStreamCodec.Collection<RegistryFriendlyByteBuf, Int, IntCollection>(StreamCodecs.VAR_INT, ::IntArrayList)) {
			container.sortWithIndices(it)
		}

		fun clientInput() {
			input.accept(container.computeSortedIndices(settings.actualComparator))
		}
	}

	fun oneWayInput(allowSpectators: Boolean = false, handler: () -> Unit): PlayerInput<Nothing?> {
		return PlayerInput(StreamCodecs.NOTHING, allowSpectators) {
			handler.invoke()
		}
	}

	fun decimalInput(allowSpectators: Boolean = false, handler: (Decimal) -> Unit) = PlayerInput(StreamCodecs.DECIMAL, allowSpectators, handler)
	fun booleanInput(allowSpectators: Boolean = false, handler: (Boolean) -> Unit) = PlayerInput(StreamCodecs.BOOLEAN, allowSpectators, handler)
	fun itemInput(allowSpectators: Boolean = false, handler: (Item) -> Unit) = PlayerInput(StreamCodecs.ITEM_TYPE, allowSpectators, handler)
	fun itemStackInput(allowSpectators: Boolean = false, handler: (ItemStack) -> Unit) = PlayerInput(ItemStack.OPTIONAL_STREAM_CODEC.wrap(), allowSpectators, handler)
	fun nullableItemInput(allowSpectators: Boolean = false, handler: (Item?) -> Unit) = PlayerInput(StreamCodecs.ITEM_TYPE.nullable(), allowSpectators, handler)
	fun stringInput(allowSpectators: Boolean = false, handler: (String) -> Unit) = PlayerInput(StreamCodecs.STRING, allowSpectators, handler)
	fun floatInput(allowSpectators: Boolean = false, handler: (Float) -> Unit) = PlayerInput(StreamCodecs.FLOAT, allowSpectators, handler)
	fun intInput(allowSpectators: Boolean = false, handler: (Int) -> Unit) = PlayerInput(StreamCodecs.INT, allowSpectators, handler)

	/**
	 * hotbar + inventory + Exopack (in this order)
	 */
	val playerInventorySlots: List<InventorySlot> = Collections.unmodifiableList(_playerInventorySlots)

	/**
	 * hotbar only
	 */
	val playerHotbarSlots: List<InventorySlot> = Collections.unmodifiableList(_playerHotbarSlots)

	/**
	 * inventory + Exopack (in this order)
	 */
	val playerCombinedInventorySlots: List<InventorySlot> = Collections.unmodifiableList(_playerCombinedInventorySlots)

	val exopackChargeSlots: List<MatteryMenuSlot> = Collections.unmodifiableList(_exopackChargeSlots)

	val exopackPowerLevel = ProfiledLevelGaugeWidget<ProfiledEnergyStorage<*>>(mSynchronizer, player.matteryPlayer.exopackEnergy)

	var sortInventoryInput: SortInput? = null
		private set

	val playerSortSettings = IItemStackSortingSettings.inputs(this, player.matteryPlayer.sortingSettings)
	val acceptExopackChargeFromWirelessCharger = BooleanInputWithFeedback(this, true, player.matteryPlayer::acceptExopackChargeFromWirelessCharger)

	var offhandSlot: InventorySlot? = null
		protected set

	var autoCreateInventoryFrame = true
		private set

	protected val lockedInventorySlots: MutableSet<Int> = HashSet()
	protected open fun isInventorySlotLocked(index: Int): Boolean = lockedInventorySlots.contains(index)

	protected var inventorySlotIndexStart = 0
	protected var inventorySlotIndexEnd = 0

	open inner class InventorySlot(container: IEnhancedContainer<IPlayerInventorySlot>, index: Int) : UserFilteredMenuSlot(container, index, 0, 0) {
		override fun mayPlace(itemStack: ItemStack): Boolean {
			return !isInventorySlotLocked(index) && super.mayPlace(itemStack)
		}

		override fun mayPickup(player: Player): Boolean {
			return !isInventorySlotLocked(index) && super.mayPickup(player)
		}

		var chargeFlag: Delegate<Boolean>? = null
			private set

		init {
			if (player.matteryPlayer.hasExopack) {
				val slot = container.containerSlot(index)
				chargeFlag = Delegate.Of(slot::shouldCharge)
			}
		}
	}

	open inner class EquipmentMenuSlot(container: IEnhancedContainer<IPlayerInventorySlot>, index: Int, val type: net.minecraft.world.entity.EquipmentSlot) : InventorySlot(container, index) {
		constructor(type: net.minecraft.world.entity.EquipmentSlot) : this(player.matteryPlayer.wrappedInventory, 34 + type.ordinal, type)

		override fun setByPlayer(newItem: ItemStack, oldItem: ItemStack) {
			inventory.player.onEquipItem(type, oldItem, newItem)
			super.setByPlayer(newItem, oldItem)
		}

		override fun mayPlace(itemStack: ItemStack): Boolean {
			return super.mayPlace(itemStack) && itemStack.canEquip(type, inventory.player)
		}

		override fun mayPickup(player: Player): Boolean {
			return super.mayPickup(player) && !item.isEmpty && (player.isCreative || !EnchantmentHelper.has(item, EnchantmentEffectComponents.PREVENT_EQUIPMENT_DROP))
		}

		override fun getMaxStackSize(): Int {
			return 1
		}

		override fun getNoItemIcon(): Pair<ResourceLocation, ResourceLocation>? {
			return Pair.of(InventoryMenu.BLOCK_ATLAS, TEXTURE_EMPTY_SLOTS[type.index])
		}
	}

	protected fun addInventorySlots(autoFrame: Boolean = !player.isSpectator) {
		check(_playerInventorySlots.isEmpty()) { "Already created inventory slots" }
		val mattery = player.matteryPlayer

		autoCreateInventoryFrame = autoFrame

		offhandSlot = object : InventorySlot(player.matteryPlayer.wrappedInventory, 40) {
			override fun setByPlayer(newItem: ItemStack, oldItem: ItemStack) {
				inventory.player.onEquipItem(net.minecraft.world.entity.EquipmentSlot.OFFHAND, oldItem, newItem)
				super.setByPlayer(newItem, oldItem)
			}

			override fun getNoItemIcon(): Pair<ResourceLocation, ResourceLocation> {
				return Pair.of(InventoryMenu.BLOCK_ATLAS, InventoryMenu.EMPTY_ARMOR_SLOT_SHIELD)
			}
		}

		mapQuickMoveToInventory(offhandSlot!!)
		addSlot(offhandSlot!!)

		for (i in 0 until if (mattery.hasExopack) mattery.combinedInventory.containerSize else mattery.wrappedItemInventory.containerSize) {
			if (i in Inventory.INVENTORY_SIZE until player.inventory.containerSize) continue

			val slot = InventorySlot(mattery.combinedInventory, i)

			_playerInventorySlots.add(slot)

			if (i <= 8)
				_playerHotbarSlots.add(slot)
			else
				_playerCombinedInventorySlots.add(slot)

			mapQuickMove(slot, equipmentSlots)
			mapQuickMoveToExternal(slot)
			addSlot(slot)
		}

		if (mattery.hasExopack) {
			_exopackChargeSlots.add(BatteryMenuSlot(mattery.exopackEnergy.parent, 0, direction = FlowDirection.OUTPUT).also { mapQuickMoveToExternal(it); mapQuickMoveToInventory(it); addSlot(it) })

			for (i in 0 until mattery.exopackChargeSlots.containerSize)
				_exopackChargeSlots.add(BatteryMenuSlot(mattery.exopackChargeSlots, i, direction = FlowDirection.INPUT).also { mapQuickMoveToExternal(it); mapQuickMoveToInventory(it); addSlot(it) })
		}

		sortInventoryInput = SortInput(mattery.inventoryAndExopackNoHotbar, playerSortSettings)
	}

	private var broadcastOnce = false

	protected open fun beforeBroadcast() {

	}

	private fun matteryBroadcast() {
		beforeBroadcast()

		mSynchronizer.observe()
		val payload = synchronizerRemote.write(player.registryAccess())

		if (payload != null) {
			PacketDistributor.sendToPlayer(player as ServerPlayer, MenuDataPacket(containerId, payload))
		}

		broadcastOnce = true
	}

	override fun broadcastChanges() {
		// how did mojang missed this
		if (player is ServerPlayer) {
			super.broadcastChanges()
			matteryBroadcast()
		}
	}

	override fun broadcastFullState() {
		// how did mojang missed this
		if (player is ServerPlayer) {
			synchronizerRemote.invalidate()
			super.broadcastFullState()
			matteryBroadcast()
		}
	}

	override fun stillValid(player: Player): Boolean {
		if (tile == null) return true

		if (player.level().getBlockEntity(tile.blockPos) !== tile) {
			return false
		}

		val pos = tile.blockPos

		return player.distanceToSqr(pos.x.toDouble() + 0.5, pos.y.toDouble() + 0.5, pos.z.toDouble() + 0.5) <= 64.0
	}

	fun syncCarried() {
		setRemoteCarried(carried.copy())
		PacketDistributor.sendToPlayer(player as ServerPlayer, SetCarriedPacket(carried))
	}

	fun syncCarried(stack: ItemStack) {
		carried = stack
		syncCarried()
	}

	private val externalSlots = ConditionalSet<Slot>()
	private val quickMoveMapping = Reference2ObjectOpenHashMap<Slot, ReferenceArrayList<Collection<Slot>>>()
	private val seenSlots = ReferenceOpenHashSet<Slot>()

	override fun addSlot(pSlot: Slot): Slot {
		if (!seenSlots.add(pSlot))
			return pSlot

		if (pSlot is MatteryMenuSlot) {
			pSlot.setupNetworkControls(this)
		}

		return super.addSlot(pSlot)
	}

	fun addSlot(slots: Iterable<Slot>) {
		slots.forEach(::addSlot)
	}

	/**
	 * Adds slot to "storage slots" list (utilized by quick move) and calls [addSlot]
	 *
	 * [condition] allows to specify when slot is invisible on GUI (hence being ignored by quick move)
	 */
	fun <T : Slot> addStorageSlot(slot: T, addMapping: Boolean = true, prepend: Boolean = false, condition: BooleanSupplier = BooleanSupplier { true }): T {
		if (!externalSlots.actuallyContains(slot)) {
			addSlot(slot)

			if (prepend)
				externalSlots.replaceFirst(slot, condition)
			else
				externalSlots.replace(slot, condition)

			if (addMapping) {
				mapQuickMove(slot, equipmentSlots)
				mapQuickMoveToInventory(slot)
			}
		}

		return slot
	}

	fun addStorageSlot(slot: Iterable<Slot>, addMapping: Boolean = true, prepend: Boolean = false, condition: BooleanSupplier = BooleanSupplier { true }) {
		for (value in slot)
			addStorageSlot(value, addMapping, prepend, condition)
	}

	/**
	 * Marks slot as "storage" - shift clicking it will move its contents to Player's inventory
	 */
	fun mapQuickMoveToInventory(slot: Slot, prepend: Boolean = false) {
		mapQuickMove(slot, playerInventorySlots, prepend = prepend)
	}

	/**
	 * Marks slot as "inventory" - shift clicking it will move its contents to menu's storage slots
	 */
	fun mapQuickMoveToExternal(slot: Slot, prepend: Boolean = false) {
		mapQuickMove(slot, externalSlots, prepend = prepend)
	}

	fun mapQuickMove(slot: Slot, target: Collection<Slot>, prepend: Boolean = false, condition: BooleanSupplier = BooleanSupplier { true }) {
		val listing = quickMoveMapping.computeIfAbsent(slot, Reference2ObjectFunction { ReferenceArrayList(1) /* ReferenceArrayList ибо мы используем его в некотором смысле как множество */ })
		listing.remove(target)

		if (prepend)
			listing.add(0, target)
		else
			listing.add(target)
	}

	fun mapQuickMove(slot: Slot, vararg target: Collection<Slot>) {
		for (value in target) {
			mapQuickMove(slot, value)
		}
	}

	// This method receive Player interactor and slot_index where Shift + Right click occurred
	// It shall return item stack that got moved
	override fun quickMoveStack(ply: Player, slotIndex: Int): ItemStack {
		val slot = slots[slotIndex]

		if (!slot.hasItem() || !slot.mayPickup(ply)) {
			return ItemStack.EMPTY
		}

		val target = quickMoveMapping[slot] ?: return ItemStack.EMPTY
		val copy = slot.item.copy()
		var any = false

		if (target.any { it.any { it.containerSlotOrNull() is IFilteredContainerSlot } }) {
			for (collection in target) {
				if (moveItemStackTo(slot, collection, onlyFiltered = true)) {
					any = true

					if (!slot.hasItem()) {
						return copy
					}
				}
			}
		}

		for (collection in target) {
			if (moveItemStackTo(slot, collection)) {
				any = true

				if (!slot.hasItem()) {
					return copy
				}
			}
		}

		if (any) {
			return copy
		} else {
			return ItemStack.EMPTY
		}
	}

	fun quickMoveToInventory(stack: ItemStack, simulate: Boolean): ItemStack {
		if (_playerInventorySlots.isEmpty()) {
			return stack
		}

		return moveItemStackToSlots(stack, _playerInventorySlots, simulate = simulate)
	}

	override fun canTakeItemForPickAll(itemStack: ItemStack, slot: Slot): Boolean {
		if (slot is EquipmentMenuSlot)
			return false

		return super.canTakeItemForPickAll(itemStack, slot) && (slot !is MatteryMenuSlot || slot.canTakeItemForPickAll()) && !slot.isCurioSlot
	}

	override fun moveItemStackTo(
		item: ItemStack,
		initialSlot: Int,
		finalSlot: Int,
		inverse: Boolean
	): Boolean {
		val remainder = moveItemStackToSlots(item, initialSlot, finalSlot, inverse)

		if (remainder.count == item.count) {
			return false
		}

		item.count = remainder.count
		return true
	}

	fun moveItemStackTo(
		source: Slot,
		slots: Collection<Slot>,
		onlyFiltered: Boolean = false
	): Boolean {
		val remainder = moveItemStackToSlots(source.item, slots, onlyFiltered = onlyFiltered)

		if (remainder.count == source.item.count) {
			return false
		}

		val copy = source.item.copy()

		if (remainder.isEmpty) {
			source.setByPlayer(ItemStack.EMPTY)
			source.onTake(player, copy)
		} else {
			copy.count = source.item.count - remainder.count
			source.item.count = remainder.count
			source.onTake(player, copy)
		}

		return true
	}

	fun moveItemStackToSlots(item: ItemStack, slots: Collection<Slot>, simulate: Boolean = false, onlyFiltered: Boolean = false): ItemStack {
		if (item.isEmpty) {
			return ItemStack.EMPTY
		}

		val copy = item.copy()

		// first pass - stack with existing slots
		if (copy.isStackable) {
			for (slot in slots) {
				if (onlyFiltered && slot.containerSlotOrNull().let { it !is IFilteredContainerSlot || it.filter == null || !it.testSlotFilter(item) }) {
					continue
				} else if (!onlyFiltered && slot.containerSlotOrNull().let { it is IFilteredContainerSlot && it.filter != null }) {
					continue
				}

				val limit = slot.getMaxStackSize(copy)

				if (limit > slot.item.count && slot.mayPlace(item) && ItemStack.isSameItemSameComponents(slot.item, copy)) {
					val newCount = (slot.item.count + copy.count).coerceAtMost(limit)
					val diff = newCount - slot.item.count
					copy.count -= diff

					if (!simulate) {
						slot.item.count += diff
						slot.setChanged()
					}

					if (copy.isEmpty) {
						return copy
					}
				}
			}
		}

		// second pass - drop stack into first free slot
		for (slot in slots) {
			if (onlyFiltered && slot.containerSlotOrNull().let { it !is IFilteredContainerSlot || it.filter == null || !it.testSlotFilter(item) }) {
				continue
			} else if (!onlyFiltered && slot.containerSlotOrNull().let { it is IFilteredContainerSlot && it.filter != null }) {
				continue
			}

			val limit = slot.getMaxStackSize(copy)

			if (!slot.hasItem() && slot.mayPlace(item)) {
				val newCount = copy.count.coerceAtMost(limit)

				if (!simulate) {
					slot.setByPlayer(copy.copy().also { it.count = newCount })
					slot.setChanged()
				}

				copy.count -= newCount

				if (copy.isEmpty) {
					return copy
				}
			}
		}

		return copy
	}

	fun moveItemStackToSlots(item: ItemStack, initialSlot: Int, finalSlot: Int, inverse: Boolean = false, simulate: Boolean = false): ItemStack {
		if (initialSlot > finalSlot) {
			return item
		}

		require(initialSlot >= 0) { "Invalid initial slot $initialSlot" }
		require(finalSlot < slots.size) { "Final slot $finalSlot is bigger than total size of array of ${slots.size}" }

		val slots = ArrayList<Slot>(finalSlot - initialSlot + 1)
		var filters = false

		for (i in (if (inverse) finalSlot downTo initialSlot else initialSlot .. finalSlot)) {
			val slot = slots[i]
			slots.add(slot)

			if (slot.containerSlotOrNull() is IFilteredContainerSlot) {
				filters = true
			}
		}

		if (filters) {
			return moveItemStackToSlots(moveItemStackToSlots(item, slots, simulate, onlyFiltered = true), slots, simulate, onlyFiltered = false)
		}

		return moveItemStackToSlots(item, slots, simulate)
	}

	private var armorSlots: ImmutableList<PlayerSlot<EquipmentMenuSlot, Slot>>? = null
	private var curiosSlots: ImmutableList<PlayerSlot<Slot, Slot>>? = null
	private val equipmentSlots = ArrayList<Slot>()

	fun makeArmorSlots(mapMoveToExternal: Boolean = false): List<PlayerSlot<EquipmentMenuSlot, Slot>> {
		if (armorSlots != null) {
			return armorSlots!!
		}

		val cosmetic = player.cosmeticArmorSlots

		return ImmutableList.of(
			PlayerSlot(EquipmentMenuSlot(net.minecraft.world.entity.EquipmentSlot.HEAD), cosmetic?.get(net.minecraft.world.entity.EquipmentSlot.HEAD)),
			PlayerSlot(EquipmentMenuSlot(net.minecraft.world.entity.EquipmentSlot.CHEST), cosmetic?.get(net.minecraft.world.entity.EquipmentSlot.CHEST)),
			PlayerSlot(EquipmentMenuSlot(net.minecraft.world.entity.EquipmentSlot.LEGS), cosmetic?.get(net.minecraft.world.entity.EquipmentSlot.LEGS)),
			PlayerSlot(EquipmentMenuSlot(net.minecraft.world.entity.EquipmentSlot.FEET), cosmetic?.get(net.minecraft.world.entity.EquipmentSlot.FEET)),
		).also {
			armorSlots = it

			for ((a, b) in it) {
				equipmentSlots.add(a)
				addSlot(a)
				if (b != null) addSlot(b)

				if (mapMoveToExternal) {
					mapQuickMoveToExternal(a)
				}

				mapQuickMoveToInventory(a)

				if (b != null) {
					if (mapMoveToExternal) {
						mapQuickMoveToExternal(b)
					}

					mapQuickMoveToInventory(b)
				}
			}
		}
	}

	fun makeEquipmentSlots(mapMoveToExternal: Boolean = false): EquipmentSlots {
		return EquipmentSlots(
			armorSlots = makeArmorSlots(mapMoveToExternal),
			curiosSlots = curiosSlots ?: ImmutableList.copyOf(player.curiosSlots).also {
				for ((a, b) in it) {
					equipmentSlots.add(a)
					addSlot(a)
					if (b != null) addSlot(b)

					if (mapMoveToExternal) {
						mapQuickMoveToExternal(a)
					}

					mapQuickMoveToInventory(a)

					if (b != null) {
						if (mapMoveToExternal) {
							mapQuickMoveToExternal(b)
						}

						mapQuickMoveToInventory(b)
					}
				}
			}.also { curiosSlots = it }
		)
	}

	companion object {
		val TEXTURE_EMPTY_SLOTS: List<ResourceLocation> = ImmutableList.of(
			InventoryMenu.EMPTY_ARMOR_SLOT_BOOTS,
			InventoryMenu.EMPTY_ARMOR_SLOT_LEGGINGS,
			InventoryMenu.EMPTY_ARMOR_SLOT_CHESTPLATE,
			InventoryMenu.EMPTY_ARMOR_SLOT_HELMET)
	}
}
